#include "global.h"
#include "flags.h"
#include "connection.h"
#include "player.h"
#include "parse.h"

// basic libs

#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

// socket related stuff

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>

// inet related stuff

#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

// threads related stuff

#include <pthread.h>

// really ugly hack

void *Start_Routine(void *);

Connection::Connection(Global *ppGlobal) {

	Connection *pWork;

	fd = 0;
	pGlobal = ppGlobal;
        id = pGlobal->GetNewConnectId();
        pPlayer = NULL;

        if(!pGlobal->GetConnectionHead())
                pGlobal->SetConnectionHead(this);
	
	pWork = pGlobal->GetConnectionHead();

	// this isn't the most efficient way to do this but it works

	while(pWork->pNext) pWork = pWork->pNext;

	pWork->pNext = this;

	if(pWork != this) pPrev = pWork;
		else pPrev = NULL;

        pNext = NULL;
}

void Connection::StartThread() {
	//todo: actual talker code goes here
	pthread_create(&thread_handle,NULL,&Start_Routine,this);

	return;
}

void Connection::StartFunction() {
	char *c;
	char username[40]; // total hack 
	ParseToken *pt;
	ParseLine *p;
	bool alive = TRUE;
	int i;

	Write("Enter your userid: ");
	c = ReadLine();

	p = new ParseLine;
	p->Set(c);
	pt = p->Parse();

	pt = pt->GetTokenNumber(0);
	
	WriteLine("Hello [%s]",pt->Get());
	strncpy(username, pt->Get(), 40); // total hack for testing/demo
	SetFlag(CFLAG_PLAYER);
	pt->ReleaseAll();

	while(alive) {
		c = ReadLine();
		p->Set(c);
		pt = p->Parse();
		pt = pt->GetTokenNumber(0);

		if(!strncmp(toupper(pt->Get()),"QUIT",4))
			Hangup(); // terminating call
		
		i = pGlobal->WriteAllPlayersExcept(this, "%s: %s\n",username,c);
		WriteLine("Okay, wrote your message to %d players",i);

	}

	Close();
}

void *Start_Routine(void *pvContext) {
	Connection *pContext = (Connection *) pvContext;
	pContext->StartFunction();
}

Connection::Connection() {
	class Connection *pWork;
	// NOTE: this code makes the connection objects nonreusable
	// because it uses a value contained in the globals

	fd = 0;
	id = pGlobal->GetNewConnectId();
	pPlayer = NULL;

	if(!pGlobal->GetConnectionHead()) 
		pGlobal->SetConnectionHead(this);
	
	pPrev = pGlobal->GetConnectionTail();
	if(pPrev) pPrev->pNext = this;
	pNext = NULL;
}

void Connection::SetGlobal(Global *ppGlobal) {
	pGlobal = ppGlobal;
}

char *Connection::ReadLine() {
	// read a line of input
	int n;
	int i = 0;

	while(1) {
		WaitForRead(); // wait to go out of blocking
		n = read(fd,&inbuf[i],INBUF_SIZE - i);

		if(!n) {
			Hangup();
		}

		inbuf[i+n] = 0;
		i+=n;
		if(strchr(&inbuf[0],'\n')) return(&inbuf[0]);
			//todo: error checking
	}

}	

bool Connection::WriteLine(char *fmt, ...) {
	// I can get away with doing things like this
	// because this is currently a multithreaded talker
	// Tears, take note - you can't get away with this shit in
	// single threaded modes

	char outbuf[1024]; // todo: bounds checking
	va_list ap;

	va_start(ap,fmt);	
	vsprintf(outbuf,fmt,ap);
	va_end(ap);
	
	strcat(outbuf,"\n"); // todo: support their term with appropriate CRLF

	write(fd,&outbuf,strlen(outbuf));
	
	return(TRUE); // todo: check for incomplete writes
	
}

bool Connection::Write(char *fmt, ...) {
	// I can get away with doing things like this
	// because this is currently a multithreaded talker
	// Tears, take note - you can't get away with this shit in
	// single threaded modes

	char outbuf[1024]; // todo: bounds checking
	va_list ap;

	va_start(ap,fmt);	
	vsprintf(outbuf,fmt,ap);
	va_end(ap);
	
	write(fd,&outbuf,strlen(outbuf));
	
	return(TRUE); // todo: check for incomplete writes
	
}

Connection::Connection(int ifd) {
	class Connection *pWork;
	
	fd = ifd;
	id = pGlobal->GetNewConnectId();
	pPlayer = NULL;

	if(!pGlobal->GetConnectionHead())
		pGlobal->SetConnectionHead(this);

	pPrev = pGlobal->GetConnectionTail();
	if(pPrev) pPrev->pNext = this;
	pNext = NULL;
}

void Connection::SetFd(int ifd) {
	fd = ifd;
}

int Connection::GetFd() {
	return(fd);
}

Connection::~Connection() {
	//fixme: This should include closing the socket, etc
	if(pPlayer) delete pPlayer;
	CloseFd();
}

bool Connection::SetFlag(long lp_Flag) {
	//note: this is NOT the same as SetFlags (!)
	// this version only allows setting one flag at a time

	if(!(l_Flags & lp_Flag)) {
		l_Flags += lp_Flag;
		return TRUE;
	} else return FALSE;

}

bool Connection::ClearFlag(long lp_Flag) {
	//note: this is NOT the same as ClearFlags (!)
	
	if(l_Flags & lp_Flag) {
		l_Flags -= lp_Flag;
		return TRUE;
	} else return FALSE;
}

bool Connection::GetFlag(long lp_Flag) {
	if(l_Flags & lp_Flag) 
		return TRUE;
	else
		return FALSE;

}

bool Connection::IsReadable() {
	fd_set fs;
	struct	timeval t;

	
	bzero(&t,sizeof(struct timeval));

	FD_ZERO(&fs);
	FD_SET(fd,&fs);

	select(fd+1,&fs,NULL,NULL,&t);

	return(FD_ISSET(fd,&fs));
}

bool Connection::IsWritable() {
	fd_set fs;
	struct	timeval t;

	bzero(&t,sizeof(struct timeval));

	FD_ZERO(&fs);
	FD_SET(fd,&fs);

	select(fd+1,NULL,&fs,NULL,&t);

	return(FD_ISSET(fd,&fs));
}

bool Connection::WaitForRead() {
	fd_set fs;
	struct timeval t;

	t.tv_sec = 1;
	t.tv_usec = 100;
	
	FD_ZERO(&fs);

	while(1) {
	
		FD_SET(fd,&fs);
		
		select(fd+1,&fs,NULL,NULL,NULL); //todo: handle errors too
						// todo: support timeout
		
		if(FD_ISSET(fd,&fs)) return(TRUE);
	}	
}
	
int Connection::SetAccept(int port) {
	//set up connection acceptance on a given port
	struct protoent *proto;
	struct sockaddr_in addr;
	int flag = 1;

	
	proto = getprotobyname("tcp");
	if(SOCKET_ERROR == (fd = socket(PF_INET,SOCK_STREAM,proto->p_proto))) {
		printf("Failed to create socket\n");	
		return(-1);
	}
	
	if(setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,&flag,sizeof(flag))) {
		printf("failed to setsockopt\n");
		CloseFd();
		return(-1);
	}

	bzero(&addr,sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	if(bind(fd,(const sockaddr *) &addr,sizeof(addr))) {
		perror("failed to bind: ");
		CloseFd();
		return(-1);
	}

	if(listen(fd,5)) {
		perror("failed to listen: ");
		CloseFd();
		return(-1);
	}

	if(fcntl(fd,F_SETFL,FNDELAY)) {
		perror("failed to set non-block: ");
		CloseFd();
		return(-1);
	}
	
	// fixme: these should all close the socket
	return(0);
}

void Connection::CloseFd(void) {
	if(!fd) return;
	close(fd);
	fd = 0;
}

void Connection::Close(void) {
	shutdown(fd,2);
	CloseFd();
}

Connection *Connection::GetNextObj() {
	return(pNext);
}

void Connection::Hangup() {
	Close();
	delete this;
	pthread_exit(NULL);
}
